@@ -33,12 +33,16 @@
 		Module Dependencies: logging
 """
 
+import types
+
 try:
 	import logging
 except:
 	import DummyLogger as logging
 	
-import simpletal
+import simpletal, simpleTAL
+
+OBJ_CONV = simpleTAL.OBJ_CONV
 
 __version__ = simpletal.__version__
 
@@ -47,6 +51,12 @@ DEFAULTVALUE = "This represents a Default value."
 class PathNotFoundException (Exception):
 	pass
 	
+class ContextContentException (Exception):
+	""" This is raised when invalid content has been placed into the Context object.
+		For example using non-ascii characters instead of Unicode strings.
+	"""
+	pass
+	
 PATHNOTFOUNDEXCEPTION = PathNotFoundException()
 
 class ContextVariable:
@@ -450,14 +460,18 @@ class Context:
 							if (endPos > 0):
 								path = expr[position + 2:endPos]
 								# Evaluate the path - missing paths raise exceptions as normal.
-								pathResult = self.evaluate (path)
+								try:
+									pathResult = self.evaluate (path)
+								except PathNotFoundException, e:
+									# This part of the path didn't evaluate to anything - leave blank
+									pathResult = u''
 								if (pathResult is not None):
-									if (type (pathResult) == type (u"")):
+									if (isinstance (pathResult, types.UnicodeType)):
 										result += pathResult
-									elif (type (pathResult) == type ("")):
-										result += unicode (pathResult, 'ascii')
 									else:
-										result += unicode (str (pathResult), 'ascii')
+										# THIS IS NOT A BUG!
+										# Use Unicode in Context if you aren't using Ascii!
+										result += OBJ_CONV (pathResult)
 								skipCount = endPos - position 
 						else:
 							# It's a variable
@@ -466,14 +480,18 @@ class Context:
 								endPos = len (expr)
 							path = expr [position + 1:endPos]
 							# Evaluate the variable - missing paths raise exceptions as normal.
-							pathResult = self.traversePath (path)
+							try:
+								pathResult = self.traversePath (path)
+							except PathNotFoundException, e:
+								# This part of the path didn't evaluate to anything - leave blank
+								pathResult = u''
 							if (pathResult is not None):
-								if (type (pathResult) == type (u"")):
+								if (isinstance (pathResult, types.UnicodeType)):
 										result += pathResult
-								elif (type (pathResult) == type ("")):
-									result += unicode (pathResult, 'ascii')
 								else:
-									result += unicode (str (pathResult), 'ascii')
+									# THIS IS NOT A BUG!
+									# Use Unicode in Context if you aren't using Ascii!
+									result += OBJ_CONV (pathResult)
 							skipCount = endPos - position - 1
 					except IndexError, e:
 						# Trailing $ sign - just suppress it
